package info.textgrid.middleware;

import java.io.IOException;
import java.io.StringWriter;
import java.math.BigInteger;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TimeZone;
import java.util.logging.Logger;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;
import org.elasticsearch.action.get.GetRequest;
import org.elasticsearch.action.get.GetResponse;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.search.fetch.subphase.FetchSourceContext;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import info.textgrid.clients.AuthClient;
import info.textgrid.clients.tgauth.AuthClientException;
import info.textgrid.middleware.oaipmh.GetRecordType;
import info.textgrid.middleware.oaipmh.HeaderType;
import info.textgrid.middleware.oaipmh.ListMetadataFormatsType;
import info.textgrid.middleware.oaipmh.MetadataFormatType;
import info.textgrid.middleware.oaipmh.MetadataType;
import info.textgrid.middleware.oaipmh.RecordType;
import info.textgrid.middleware.oaipmh.Resource;
import info.textgrid.middleware.oaipmh.ResumptionTokenType;
import info.textgrid.namespaces.middleware.tgauth.ProjectInfo;

/**
 * <p>
 * OAIPMH Utilities is responsible to implement all necessary functions to implement the arguments
 * of a request.
 * </p>
 * 
 * @author Maximilian Brodhun, SUB Göttingen
 * @author Stefan E. Funk, SUB Göttingen
 */
public class OaipmhUtilities {

  private static Logger log = Logger.getLogger(OaipmhUtilities.class.getName());

  // **
  // FINALS
  // **

  public static final String ES_DIVIDER_CHAR = ".";
  public static final String ES_DIVIDER_REGEXP = "\\.";

  public static final String PREFIX_DEVIDER = ":";
  public static final String PREFIX_HDL = "hdl";
  public static final String PREFIX_TEXTGRID = "textgrid";
  public static final Set<String> IDENTIFIER_PREFIXES_TO_FILTER = new HashSet<String>(Arrays.asList(
      PREFIX_HDL,
      PREFIX_TEXTGRID));

  // **
  // CLASS
  // **

  private String rangeField;
  private long resultSize;
  private boolean foundItems;
  private String formatField;
  private ResumptionTokenType resTokenForResponse;
  private String formatToFilter;
  private int searchResponseSize;

  // **
  // STATICS
  // **

  /**
   * @param verb
   * @return
   */
  public static boolean isOAIPMH_RequestArgument(String verb) {
    String allowedArgument = "^[identifier,from,until,set,resumptionToken,metadataPrefix]*$";
    return verb.matches(allowedArgument);
  }

  /**
   * <p>
   * Since TextGrid just supports DC the only returned MetadataFormat is DublinCore.
   * </p>
   * 
   * @return
   */
  public MetadataFormatType setMetadataFormats() {

    MetadataFormatType tgDublinCore = new MetadataFormatType();
    tgDublinCore.setMetadataNamespace("http://www.openarchives.org/OAI/2.0/oai_dc/");
    tgDublinCore.setMetadataPrefix("oai_dc");
    tgDublinCore.setSchema("http://www.openarchives.org/OAI/2.0/oai_dc.xsd");

    return tgDublinCore;
  }

  /**
   * @return
   * @throws AuthClientException
   */
  public static List<ProjectInfo> getProjectList() throws AuthClientException {

    AuthClient auth = new AuthClient();
    List<ProjectInfo> projectInfos = auth.getAllProjects();
    auth.getAllProjects();

    return projectInfos;
  }

  /**
   * @param projectID
   * @return
   * @throws AuthClientException
   */
  public static String getProjectName(String projectID) throws AuthClientException {
    AuthClient auth = new AuthClient();
    return auth.getProjectInfo(projectID).getName();
  }

  /**
   * <p>
   * Producing the list of all metadata formats.
   * </p>
   * 
   * @return tgRepMetadaFormats
   */
  public ListMetadataFormatsType setMetadataFormatList() {

    ListMetadataFormatsType tgRepMetadataFormats = new ListMetadataFormatsType();
    MetadataFormatType tgDublinCore = setMetadataFormats();
    tgRepMetadataFormats.getMetadataFormat().add(tgDublinCore);

    return tgRepMetadataFormats;
  }

  /**
   * <p>
   * Calculating the current date and give it back as XMLGregorianCalendar.
   * </p>
   * 
   * @return now
   * @throws DatatypeConfigurationException
   */
  public static XMLGregorianCalendar getXMLGregorianCalendarNow()
      throws DatatypeConfigurationException {

    GregorianCalendar gregorianCalendar = new GregorianCalendar();
    DatatypeFactory datatypeFactory = DatatypeFactory.newInstance();
    XMLGregorianCalendar now = datatypeFactory.newXMLGregorianCalendar(gregorianCalendar);

    return now;
  }

  /**
   * <p>
   * Converting a given string representing a date value into the date format required for OAIPMH
   * and give it back as XMLGregorianCalendar.
   * </p>
   * 
   * @param originalDateTimeString
   * @return xmlCal
   * @throws ParseException
   * @throws DatatypeConfigurationException
   */
  public static XMLGregorianCalendar oaiDatestampAsGregorian(String originalDateTimeString)
      throws ParseException, DatatypeConfigurationException {

    log.fine("original date time string: " + originalDateTimeString);

    SimpleDateFormat tgItemTime = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.S");
    Date date = tgItemTime.parse(originalDateTimeString);
    SimpleDateFormat outFormatter = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");
    outFormatter.setTimeZone(TimeZone.getTimeZone("UTC"));
    String output2 = outFormatter.format(date);
    XMLGregorianCalendar xmlCal = DatatypeFactory.newInstance().newXMLGregorianCalendar(output2);

    return xmlCal;
  }

  /**
   * <p>
   * Converting a given string representing a date value into the date format required for OAIPMH
   * and give it back as string.
   * </p>
   * 
   * @param originalDateTimeString
   * @return dateOutputAsString
   * @throws ParseException
   */
  public static String oaiDatestampAsString(String originalDateTimeString) throws ParseException {

    log.fine("incoming date: " + originalDateTimeString);

    SimpleDateFormat tgItemTime = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.S");
    Date date = tgItemTime.parse(originalDateTimeString);
    SimpleDateFormat outFormatter = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");
    outFormatter.setTimeZone(TimeZone.getTimeZone("UTC"));
    String dateOutputAsString = outFormatter.format(date);

    log.fine("outgoing date: " + dateOutputAsString);

    return dateOutputAsString;
  }

  /**
   * <p>
   * Returns a resumptionToken for a search, if necessary.
   * </p>
   * 
   * TODO Check if we really must put all these references in!
   * 
   * @param completeListSize
   * @param resumptionToken
   * @param cursorCollector
   * @param scrollID
   * @param searchResponseSize
   * @param i
   * @return A resumption token if applicable, null otherwise.
   */
  public static ResumptionTokenType getResumptionToken(final long completeListSize,
      final String resumptionToken, Map<String, Integer> cursorCollector, final String scrollID,
      final int searchResponseSize, final int i) {

    log.fine("Creating a ResumptionToken:\n" + "  CompleteListSize: " + completeListSize + "\n"
        + "  ResumptionToken: " + resumptionToken + "\n" + "  CursorCollector: " + cursorCollector
        + "\n" + "  ScrollID: " + scrollID + "\n" + "  SearchResponseSize: " + searchResponseSize
        + "\n" + "  IterationStep: " + i);

    int cursor;

    // Three cases here:
    // 1. Complete list size is > searchResponseSize and a token already is existing: we need to
    // check hash map!
    // 2. Complete list size is > searchResponseSize and we have no token: we do need one!
    // 3. Complete list size is <= searchResponseSize (we do not need a token! do nothing!)
    if (completeListSize > searchResponseSize) {
      ResumptionTokenType resTokenForResponse = new ResumptionTokenType();

      if (resumptionToken != null && cursorCollector.containsKey(resumptionToken)) {

        cursor = cursorCollector.get(resumptionToken).intValue() + i;
        if (cursor == searchResponseSize) {
          // TODO Why not "cursor += searchResponseSize;" ??
          cursor = 2 * searchResponseSize;
        }
        resTokenForResponse.setCursor(BigInteger.valueOf((long) cursor));
        cursorCollector.put(scrollID, cursor);
      } else {
        resTokenForResponse.setCursor(BigInteger.valueOf(searchResponseSize));
        cursorCollector.put(scrollID, searchResponseSize);
        cursor = searchResponseSize;
        resTokenForResponse.setCursor(BigInteger.valueOf((long) cursor));

        log.fine("New ResumptionToken created from scroll ID: " + cursorCollector.get(scrollID));
      }

      // Set resumption token string if cursor is less then complete list size: More objects can be
      // delivered!
      if (cursor < completeListSize) {
        resTokenForResponse.setValue(scrollID);
      }

      // Remove resumption token string if cursor is bigger then or equal complete list size: This
      // is the last response, no more objects are available!
      if (cursor >= completeListSize) {
        resTokenForResponse.setValue("");
      }

      // Set complete list size.
      resTokenForResponse.setCompleteListSize(BigInteger.valueOf(completeListSize));

      return resTokenForResponse;
    }

    return null;
  }

  /**
   * @param dateToValidate
   * @return
   */
  public static boolean isThisDateValid(String dateToValidate) {

    SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.S");

    dateFormat.setLenient(false);
    try {
      dateFormat.parse(dateToValidate.trim());
    } catch (ParseException pe) {
      return false;
    }
    return true;
  }

  /**
   * @param dateToValidate
   * @return
   */
  public static boolean isThisDateValidToOtherTimeStamp(String dateToValidate) {

    SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");

    dateFormat.setLenient(false);
    dateFormat.setLenient(false);
    try {
      dateFormat.parse(dateToValidate.trim());
    } catch (ParseException pe) {
      return false;
    }
    return true;
  }

  /**
   * 
   */
  private static final String[] formats = {
      "yyyy-MM-dd'T'HH:mm:ss'Z'", "yyyy-MM-dd'T'HH:mm:ssZ",
      "yyyy-MM-dd'T'HH:mm:ss", "yyyy-MM-dd'T'HH:mm:ss.SSS'Z'",
      "yyyy-MM-dd'T'HH:mm:ss.SSSZ", "yyyy-MM-dd HH:mm:ss",
      "MM/dd/yyyy HH:mm:ss", "MM/dd/yyyy'T'HH:mm:ss.SSS'Z'",
      "MM/dd/yyyy'T'HH:mm:ss.SSSZ", "MM/dd/yyyy'T'HH:mm:ss.SSS",
      "MM/dd/yyyy'T'HH:mm:ssZ", "MM/dd/yyyy'T'HH:mm:ss",
      "yyyy:MM:dd HH:mm:ss", "yyyyMMdd",};

  /**
   * @param d
   * @return
   */
  public static String getFormatOfDate(String d) {

    if (d != null) {
      for (String parse : formats) {
        SimpleDateFormat sdf = new SimpleDateFormat(parse);
        try {
          sdf.parse(d);
          return parse;
        } catch (ParseException e) {
          // TODO Handle Exception!
        }
      }
    }

    return d;
  }

  /**
   * <p>
   * Loops over all fields and collects field contents.
   * </p>
   * 
   * @param resultFromGetRequestInES
   * @param fields
   * @return Returns a list with all field entries.
   */
  public static List<String> loopListFieldLoader(final JSONObject resultFromGetRequestInES,
      final String[] fields) {

    List<String> result = new ArrayList<String>();

    for (String f : fields) {
      result.addAll(listFieldLoader(resultFromGetRequestInES, f));
    }

    return result;
  }

  /**
   * <p>
   * Loads the field with one param.
   * </p>
   * 
   * @param resultFromGetRequestInES
   * @param field
   * @return Returns one or more params as a list.
   */
  public static List<String> listFieldLoader(final JSONObject resultFromGetRequestInES,
      final String field) {

    List<String> fieldResults = new ArrayList<String>();

    log.fine("JSON: " + (resultFromGetRequestInES == null ? "NULL" : resultFromGetRequestInES));
    log.fine("FIELD: " + field);

    // Check if field has got "." in it, such as
    // "edition.source.bibliographicCitation.placeOfPublication".
    String[] fieldPathForESIndex = field.split(ES_DIVIDER_REGEXP);

    JSONObject singlePath = resultFromGetRequestInES;
    try {
      // Loop the splitted field components, such as "edition", "source", etcpp.
      for (int i = 0; i < fieldPathForESIndex.length; i++) {

        String fieldComponent = fieldPathForESIndex[i];

        log.fine("i=" + i);
        log.fine("fieldPathLength=" + fieldPathForESIndex.length);

        // Case 1: Reduce JSON object.
        if (i < fieldPathForESIndex.length - 1) {

          log.fine(
              "case 1: " + i + "<" + (fieldPathForESIndex.length - 1) + " : reduce singlePath");

          singlePath = singlePath.getJSONObject(fieldPathForESIndex[i]);
        }
        // Case 2: We have one field component.
        else if (fieldPathForESIndex.length == 1) {

          log.fine("case 2: we have only one field component in field " + field);

          // Add content from JSON, Array or String.
          fieldResults.addAll(getFieldContent(singlePath, fieldComponent));

        }
        // Case 3: We have got more field components.
        else {

          log.fine("case 3: we have " + fieldPathForESIndex.length + " components in " + field);

          // Add content from JSON, Array or String.
          fieldResults.addAll(getFieldContent(singlePath, fieldComponent));

        }
      }
    } catch (JSONException e) {
      log.fine("IGNORING JSON ERROR: " + e.getMessage());
    }

    log.fine("field results: " + fieldResults);

    return fieldResults;
  }

  /**
   * <p>
   * Loads the field with the first param from. Please use for single value fields only.
   * </p>
   * 
   * @param resultFromGetRequestInES
   * @param field
   * @return Returns only the first String from the result list, or "" if list is empty.
   */
  public static String firstEnrtryFieldLoader(JSONObject resultFromGetRequestInES, String field) {

    String result = "";

    List<String> list = listFieldLoader(resultFromGetRequestInES, field);
    if (!list.isEmpty()) {
      result = list.get(0);
    }

    return result;
  }

  /**
   * @param theESClient The ES client to use.
   * @param idInElasticSearchIndex The ID to get the ES record from.
   * @param includes Including the fields to fetch from the elastic search index.
   * @param excludes Excluding the fields to fetch from the elastic search index.
   * @return
   * @throws IOException
   */
  protected static GetResponse getRcordByIDFromElasticSearch(OaipmhElasticSearchClient theESClient,
      String idInElasticSearchIndex, String[] includes, String[] excludes) throws IOException {

    log.fine("id in es index: " + idInElasticSearchIndex);
    log.fine("include/exclude: " + includes.length + "/" + excludes.length);

    // Setting the source context for fetching the fields from the elastic search index
    FetchSourceContext fetchSourceContext = new FetchSourceContext(true, includes, excludes);

    log.fine("esclient/index: " + (theESClient != null ? theESClient : "null") + "/"
        + (theESClient != null ? theESClient.getEsIndex() : "null"));

    // Check for ES client.
    if (theESClient == null) {
      throw new IOException("elasticsearch client is NULL!");
    }

    // Building the getRequest against the elastic search index.
    GetRequest getRequest =
        new GetRequest(theESClient.getEsIndex(), theESClient.getEsType(), idInElasticSearchIndex)
            .fetchSourceContext(fetchSourceContext);

    log.fine("get request id: " + getRequest.id());

    // Declaration of the result from the get-request
    GetResponse esResultObject = theESClient.getEsClient().get(getRequest, RequestOptions.DEFAULT);

    return esResultObject;
  }

  /**
   * @param resource
   * @throws JAXBException
   */
  public static void marshal(Resource resource) throws JAXBException {
    JAXBContext jc = JAXBContext.newInstance(Resource.class);
    Marshaller marshaller = jc.createMarshaller();
    StringWriter stringWriter = new StringWriter();
    marshaller.marshal(resource, stringWriter);
    marshaller.marshal(resource, System.out);
  }

  /**
   * @param metadata
   * @param header
   * @return
   */
  public static GetRecordType getRecordType(Object metadata, HeaderType header) {

    MetadataType metadataType = new MetadataType();
    RecordType recordType = new RecordType();
    GetRecordType getRecordType = new GetRecordType();

    metadataType.setAny(metadata);
    recordType.setMetadata(metadataType);
    recordType.setHeader(header);
    getRecordType.setRecord(recordType);

    return getRecordType;
  }

  /**
   * @param propertyName
   * @param propFile
   * @return
   */
  @Deprecated
  public static String[][] fetchArrayFromPropFile(String propertyName, Properties propFile) {

    // TODO Why get a 2D array and only use the first element in RecordDelivererDatacite?
    // BTW Where does the ";" come from anyway?

    // Get array split up by the semicolon.
    String[] a = propFile.getProperty(propertyName).split(";");

    // Create the two dimensional array with correct size.
    String[][] array = new String[a.length][a.length];

    // Combine the arrays split by semicolon and comma.
    for (int i = 0; i < a.length; i++) {
      array[i] = a[i].split(",");
    }

    return array;
  }

  /**
   * @param propertyName
   * @param propFile
   * @return
   */
  public static String[] fetchFieldsFromPropFile(String propertyName, Properties propFile) {
    // Trim with whitespaces, so we get the array with results trimmed.
    return propFile.getProperty(propertyName).split("\\s*,\\s*");
  }

  /**
   * <p>
   * Check identifier for prefixes: Remove "textgrid:" and "hdl:" prefixes.
   * </p>
   * 
   * @param identifier
   * @return The identifier without prefix, if contained in defined prefix list.
   */
  public static String omitPrefixFromIdentifier(final String identifier) {

    String result = identifier;

    if (identifier != null && !identifier.isEmpty() && identifier.contains(PREFIX_DEVIDER)) {
      String prefix[] = identifier.split(PREFIX_DEVIDER);
      String px = prefix[0];
      String id = prefix[1];
      if (IDENTIFIER_PREFIXES_TO_FILTER.contains(px)) {
        result = id;
        log.fine("incoming id " + id + " has been filtered to: " + result);
      }
    }

    return result;
  }

  /**
   * <p>
   * Create header type of OAI-PMH response.
   * </p>
   * 
   * @param dateOfCreation The date of the resource's creation.
   * @param identifier The identifier to set in the header. Please see to it, that it is correct
   *        here, and have not to be changed in any way.
   * @param setSpec The ID of the set spec.
   * @return
   */
  public static HeaderType computeResponseHeader(final String dateOfCreation,
      final String identifier, final String setSpec) {

    HeaderType result = new HeaderType();

    result.setIdentifier(identifier);
    result.setDatestamp(dateOfCreation);

    log.finer(
        "Header (DoC / ID / setSpec): " + dateOfCreation + " / " + identifier + " / " + setSpec);

    if (setSpec != null && !setSpec.isEmpty()) {
      result.getSetSpec().add(setSpec);
    }

    return result;
  }

  /**
   * <p>
   * Get the setSpec, if given, check for prefix, if not given, use prefix and given ID.
   * </p>
   * 
   * @param theSetSpec
   * @param theSpecPrefix
   * @param theID
   * @return A valid setSpec value for the set.
   */
  public static String getSetSpec(String theSetSpec, String theSpecPrefix, String theID) {

    String result = theSetSpec;

    // Check given setSpec value, add ID if empty.
    if (result == null || result.isEmpty() || result.equals(theSpecPrefix)) {
      result = theID;
    }

    // Then check for prefix, add prefix, if necessary.
    if (!result.startsWith(theSpecPrefix)) {
      result = theSpecPrefix + result;
    }

    return result;
  }

  /**
   * @param theSID
   * @return
   */
  public static String hideSID(String theSID) {

    String result = "(null or empty)";

    if (theSID != null && !theSID.isEmpty()) {
      int size = theSID.length();
      if (size < 5) {
        result = "(sid existing and too short)";
      } else {
        result = theSID.substring(0, 4) + "...";
      }
    }

    return result;
  }

  /**
   * @param theURI
   * @return
   */
  public static String getTextGridBaseURI(String theURI) {

    String result = theURI;

    if (theURI.contains(".")) {
      result = theURI.substring(0, theURI.indexOf('.'));
    }

    return result.trim();
  }

  /**
   * @param theURI
   * @return
   */
  public static String getTextGridRevisionURI(String theURI) {

    // TODO Get latest revision from tgsearch?
    // https://dev.textgridlab.org/1.0/tgsearch/info/textgrid:3qmmt/revisions?sid=***

    String result = theURI;

    if (!theURI.contains(".")) {
      result = getTextGridBaseURI(theURI) + ".0";
    }

    return result.trim();
  }

  // **
  // PRIVATE METHODS
  // **

  /**
   * <p>
   * Get List of Strings from JSON object, Array or String!
   * </p>
   * 
   * @param singlePath
   * @param fieldComponent
   * @return
   */
  private static List<String> getFieldContent(JSONObject singlePath, String fieldComponent) {

    List<String> result = new ArrayList<String>();

    // Examine JSONArray, take String if not applicable.
    try {
      JSONArray singlePathArray = singlePath.getJSONArray(fieldComponent);

      log.fine("get json array: " + singlePathArray.toString(2));

      for (int j = 0; j < singlePathArray.length(); j++) {
        result.add(singlePathArray.getString(j));
      }
    } catch (JSONException e) {
      Object noArray = singlePath.get(fieldComponent);

      if (noArray instanceof String) {

        log.fine("get string: " + noArray);

        result.add(singlePath.getString(fieldComponent));

      } else if (noArray instanceof Integer) {

        log.fine("get int: " + noArray);

        result.add(String.valueOf(singlePath.getInt(fieldComponent)));
      } else {
        log.fine("no string & no int: " + noArray);
      }
    }

    return result;
  }

  // **
  // GETTERS & SETTERS
  // **

  /**
   * @return
   */
  public String getRangeField() {
    return this.rangeField;
  }

  /**
   * @param rangeField
   */
  public void setRangeField(String rangeField) {
    this.rangeField = rangeField;
  }

  /**
   * @return
   */
  public long getResultSize() {
    return this.resultSize;
  }

  /**
   * @param resultSize
   */
  public void setResultSize(long resultSize) {
    this.resultSize = resultSize;
  }

  /**
   * @return
   */
  public boolean isFoundItems() {
    return this.foundItems;
  }

  /**
   * @param foundItems
   */
  public void setFoundItems(boolean foundItems) {
    this.foundItems = foundItems;
  }

  /**
   * @return
   */
  public ResumptionTokenType getResTokenForResponse() {
    return this.resTokenForResponse;
  }

  /**
   * @param resTokenForResponse
   */
  public void setResTokenForResponse(ResumptionTokenType resTokenForResponse) {
    this.resTokenForResponse = resTokenForResponse;
  }

  /**
   * @return
   */
  public String setFormatField() {
    return this.formatField;
  }

  /**
   * @param formatField
   */
  public void setFormatField(String formatField) {
    this.formatField = formatField;
  }

  /**
   * @return
   */
  public String getFormatToFilter() {
    return this.formatToFilter;
  }

  /**
   * @param formatToFilter
   */
  public void setFormatToFilter(String formatToFilter) {
    this.formatToFilter = formatToFilter;
  }

  /**
   * @return
   */
  public int getSearchResponseSize() {
    return this.searchResponseSize;
  }

  /**
   * @param searchResponseSize
   */
  public void setSearchResponseSize(int searchResponseSize) {
    this.searchResponseSize = searchResponseSize;
  }

}
